# Preprocessors

- [Overview](#overview)
- [Purpose](#purpose)
- [Features](#features)
- [How To Use](#how-to-use)

## Overview
Recognize this simple hook function?

```php
/**
 * Implements hook_preprocess_HOOK() for page templates.
 */
function THEMENAME_preprocess_node(&$variables) {
    $variables['hello'] = 'world';
}
```

You would normally place this in your `THEMENAME.theme` file to make a
variable called `hello` accessible in all node template files. Easy!

But often times on websites your `THEMENAME.theme` becomes large and
difficult to maintain, as preprocessing is often whats used to clean up
data before it reaches templates. There are strategies around this, but
this module offers one that fits rather well with Drupal's tendencies
towards using the strength of the
[Plugin API](https://www.drupal.org/docs/8/api/plugin-api).

Introducing Preprocessors!

```php
<?php

namespace Drupal\THEME_NAME\Plugin\preprocessors;

use Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorPluginBase;

/**
 * Provides a Preprocessor for nodes.
 */
final class NodePreprocessor extends PreprocessorPluginBase {

  /**
   * {@inheritdoc}
   */
  public function preprocess(&$variables, $hook, $info) {
    $variables['hello'] = 'world';
  }

}
```

If you are familiar with the Plugin API, both
**Annotated Class Discovery** and **Yaml Discovery** are supported. Only
**Yaml Discovery** is supported in themes.

If you are unfamiliar with the Plugin API, you may read up [here](https://www.drupal.org/docs/8/api/plugin-api).
Otherwise, the guide below is pretty straightforward. There is also an
examples module provided with quick examples of how to implement
Preprocessors!

## Purpose
The ultimate goal of this module is to better organize preprocessing of
templates in Drupal.

## Advantages
- Preprocess template variables through the Plugin API.
- Use of Dependency Injection when preprocessing.
- Better file organization and OOP advantages with class inheritance.
- Better readability overall.

## Usage Guide
Normally, when attempting to preprocess the variables of a **node** for
example, you would create a function such as the following in your
THEME_NAME.theme file:

```php
/**
 * Implements hook_preprocess_HOOK() for page templates.
 */
function THEMENAME_preprocess_node(&$variables) {
    $variables['hello'] = 'world';
}
```

This module allows you to implement **Preprocessor** plugins instead.

Preprocessor plugins have 4 possible properties:
- ID **(Required)**: The unique ID of the plugin. This is usually
`MODULE_NAME.CLASS_NAME` or `THEME_NAME.CLASS_NAME`.
- Theme Hook **(Required)**: The theme hook that this plugin will apply
preprocessing for. This is equivalent to the value of `HOOK` in a
regular `hook_preprocess_HOOK()` function.
- Weight **(Optional)**: The weight relative to other preprocessors that
act upon the same template. Lower value means earlier preprocessing.
**Defaults to 0.**
- Themes **(Optional)**: The Themes in which this plugin will act. If
the plugin is provided by a theme, this property should be omitted as it
will only act in the theme its defined in. Otherwise, in modules, it is
optional, and plugins will act in all themes if this is omitted.

There are two main ways to create **preprocessors**:
- Through a theme.
- Through a module.

Below, I will include examples of how to create functionality that does
exactly the same thing as the simple hook function above, but through
Plugins instead.

### Adding preprocessors to a theme
In a theme, the first step would be to create a
`THEME_NAME.preprocessors.yml` at the root of the theme. The contents
of this file should be similar  to this:

```yaml
THEME_NAME.node_preprocessor:
  class: '\Drupal\THEME_NAME\Plugin\preprocessors\NodePreprocessor'
  theme_hook: node
  weight: 0
```

Here, you are declaring a plugin with the id
`THEME_NAME.node_preprocessor` and setting the path towards its class
that we will create shortly. The `theme_hook` key actually tells the
plugin which `theme_hook` we are preprocessing, and is equivalent to the
value you would replace `HOOK` with in any `hook_preprocess_HOOK()`
function.

Now, you can create a `NodePreprocessor.php` file.

File: `/THEME_NAME/src/Plugin/preprocessors/NodePreprocessor.php`

```php
<?php

namespace Drupal\THEME_NAME\Plugin\preprocessors;

use Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorPluginBase;

/**
 * Provides a Preprocessor for nodes.
 */
final class NodePreprocessor extends PreprocessorPluginBase {

  /**
   * {@inheritdoc}
   */
  public function preprocess(&$variables, $hook, $info) {
    $variables['hello'] = 'world';
  }

}
```

And you're done! Clear your cache, and your class should now function
as a preprocessor for nodes! Beyond this, you can use Dependency
Injection to inherit some of Drupal's services, or use more advanced OOP
patterns to create any kind of relations between multiple preprocessors
you may have.

### Adding preprocessors to a module
Adding a preprocessor to a module is pretty much the same deal as
with a theme, but there are two methods.

The first would be to use the same method as used above for the theme,
as in creating a `MODULE_NAME.preprocessors.yml` file and following the
same instructions.

Otherwise, you can also use **Class Annotations** to fulfill the same
goal. In this case, you don't need a Yaml file, and you can proceed by
simply creating the following file.

File: `/MODULE_NAME/src/Plugin/preprocessors/NodePreprocessor.php`

```php
<?php

namespace Drupal\MODULE_NAME\Plugin\preprocessors;

use Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorPluginBase;

/**
 * Provide plugin to alter variables for nodes.
 *
 * @noinspection
 *   AnnotationMissingUseInspection
 *
 * @Preprocessor(
 *   id = "preprocessors_example.node_preprocessor",
 *   theme_hook = "node",
 *   weight = 0,
 *   themes = {
 *      "YOUR_THEME"
 *   }
 * )
 *
 * @package Drupal\preprocessors_example\Plugin\Preprocessor
 */
final class NodePreprocessor extends PreprocessorPluginBase {

  /**
   * Add personal tweaks to variables in this function.
   *
   * {@inheritdoc}
   */
  public function preprocess(&$variables, $hook, $info): void {
    $variables['hello'] = 'world';
  }

}
```

As shown in the code above, you define the plugin's details in the class
comments as an Annotation, as opposed to the `.yml` file. The
properties are the same as above.

The `themes` property is more used in modules in case you
want your preprocessor to be limited to a selection of themes. If this
property is omitted, it will be applied in all themes.

## Examples
You can find simple examples in the `preprocessors_example` module found
in the `modules` folder of this module. Take a look and even copy its
structure to get started quickly.
