<?php

namespace Drupal\preprocessors\Service;

use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Theme\ThemeManager;
use Drupal\Core\Theme\Registry;
use Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorInterface;
use Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorsPluginManager;

/**
 * Provides an object to manage core functionality of the module.
 *
 * @package Drupal\preprocessors\Service
 */
final class PreprocessorsManager {

  /**
   * The Module Handler injected through DI.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  private $moduleHandler;

  /**
   * Theme Manager injected through DI.
   *
   * @var \Drupal\Core\Theme\ThemeManager
   */
  private $themeManager;

  /**
   * The Theme Registry injected through DI.
   *
   * @var \Drupal\core\Theme\Registry
   */
  private $themeRegistry;

  /**
   * Preprocessor Plugin Manager injected through DI.
   *
   * @var \Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorsPluginManager
   */
  private $preprocessorsPluginManager;

  /**
   * PreprocessorsManager constructor.
   *
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler
   *   The module handler.
   * @param \Drupal\Core\Theme\ThemeManager $themeManager
   *   Theme Manager injected through DI.
   * @param \Drupal\Core\Theme\Registry $themeRegistry
   *   The theme registry.
   * @param \Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorsPluginManager $preprocessorsPluginManager
   *   Preprocessor Plugin Manager injected through DI.
   */
  public function __construct(ModuleHandlerInterface $moduleHandler, ThemeManager $themeManager, Registry $themeRegistry, PreprocessorsPluginManager $preprocessorsPluginManager) {
    $this->moduleHandler = $moduleHandler;
    $this->themeManager = $themeManager;
    $this->themeRegistry = $themeRegistry;
    $this->preprocessorsPluginManager = $preprocessorsPluginManager;
  }

  /**
   * Get theme suggestions for a given hook.
   *
   * @param string $hook
   *   Name of the hook.
   * @param array $variables
   *   Variables to preprocess.
   *
   * @return array
   *   Hook's theme suggestions.
   */
  public function getSuggestions(string $hook, array $variables): array {
    $suggestions = $this->moduleHandler->invokeAll('theme_suggestions_' . $hook, [$variables]);
    $hooks = [
      'theme_suggestions',
      'theme_suggestions_' . $hook,
    ];
    $this->moduleHandler->alter($hooks, $suggestions, $variables, $hook);
    $this->themeManager->alter($hooks, $suggestions, $variables, $hook);

    return $suggestions;
  }

  /**
   * Preprocess variables using Preprocessor Plugins.
   *
   * @param array  $variables
   *   Array of default variables obtained from the render instance.
   * @param string $hook
   *   The name of the theme hook being invoked. This is not always accurate.
   * @param array  $info
   *   Array of information on the current render instance. This parameter is
   *   not even documented on Drupal's main site, but the module's functionality
   *   relies on it. If this breaks, everything breaks.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function preprocess(array &$variables, string $hook, $info): void {
    // Get the hook's theme suggestions.
    $suggestions = $this->getSuggestions($hook, $variables);

    // Perform preprocessing with the base hook first.
    $this->preprocessHook($variables, $hook, $info);

    // If there are no suggestions, we can exit here.
    if (empty($suggestions)) {
      return;
    }

    $theme_registry = $this->themeRegistry->getRuntime();
    foreach ($suggestions as $suggestion) {
      if (!$theme_registry->has($suggestion)) {
        continue;
      }

      $this->preprocessHook($variables, $suggestion, $info);
    }
  }

  /**
   * Handles the preprocessing of a hook.
   *
   * @param string $hook
   *   The preprocess hook.
   * @param array  $variables
   *   The variables to preprocess.
   * @param array  $info
   *   Array of information on the current render instance. This parameter is
   *   not even documented on Drupal's main site, but the module's functionality
   *   relies on it. If this breaks, everything breaks.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function preprocessHook(array &$variables, string $hook, array $info): void {
    // We get all preprocessors associated with the provided hook.
    $preprocessors = $this->preprocessorsPluginManager->getPreprocessorsForActiveTheme($hook);

    // Can't do anything without any preprocessors!
    if (empty($preprocessors)) {
      return;
    }

    // Sort the preprocessors by weight.
    uasort($preprocessors, 'self::sortPreprocessors');

    /* @var \Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorInterface $preprocessor */
    foreach ($preprocessors as $preprocessor) {
      // We want to sort them by weight.
      $preprocessor->preprocess($variables, $hook, $info);
    }
  }

  /* @noinspection PhpUnusedPrivateMethodInspection */

  /**
   * Sort preprocessors according to their weight.
   *
   * @param \Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorInterface $a
   *   First preprocessor.
   * @param \Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorInterface $b
   *   Second preprocessor.
   *
   * @return int
   *   Returns a value less than, equal to or greater than zero depending on if
   *   the first argument is considered to be respectively less than, equal to
   *   or greater than the second.
   */
  private function sortPreprocessors(PreprocessorInterface $a, PreprocessorInterface $b): int {
    return $a->getWeight() - $b->getWeight();
  }

}
